(defparameter *position* 150)
(defparameter *results* nil)
(defparameter *problem-result* nil)
(defparameter *problem-start-time* nil)
(defparameter *response* nil)
(defparameter *in-nback* nil)
(defparameter *real-clicks* nil)
(defparameter *nback-match* 0)

(defparameter *rd-letter* nil)

(defparameter *timeout-event* nil)


(defparameter *experiment-events* nil)
(defparameter *bold-predictions* nil)


(defparameter *problem* '(  8 $ 3 = 21))


(defun run-problem ()
  (reset)
  (let* ((window (open-exp-window "Pyramid Experiment" :visible nil))
        (items (list (first *problem*) (third *problem*) 'x))
        (wait-time nil))
        (princ (format nil "Starting fixation~%"))
        ; note that goal activity at fixation changes BOLD!
        (log-experiment-event 1);
        ;(run-full-time 1) ; fixation
        (setf wait-time (synchronize-fixation))
       ; (mod-focus-fct (list 'state 'encode-base 'step 'looking 'next-state 'encode-height 'role 'base 'value nil 'current-term nil 
        (schedule-event-relative wait-time (lambda () 
                                             (princ (format nil "Presenting ~A $ ~A = ~A~%" (first items) (second items) (third items)))
                                             (log-experiment-event 2)
                                             (setf *problem-result* nil)
                                             (setf *problem-start-time* (/ (get-time) 1000.0))
                                             (mod-focus-fct (list 'state 'encode-base 'step 'ready 'next-state nil 'role nil 'value nil 'current-term nil 
                                                                  'current-problem 1 'current-problem-position 'P1))
                                             (clear-buffer 'imaginal)
                                             (clear-buffer 'retrieval)
                                             (setf *position* 150)
                                             (install-device window)
                                             (present-problem items)
                                             (proc-display :clear t)
                                             (setf *real-clicks* nil)))
        (setf *timeout-event* (schedule-event-relative (+ wait-time 30) 'schedule-post-answer))
        (run 1000)
        (predict-augmented-bold-response)))


(defmethod rpm-window-key-event-handler ((win rpm-window) key)
(if (not *in-nback*)
  (progn
    (setf *response* (string key))
    (print *response*)
    (if (not (string-equal *response* "r"))
        (progn 
          (add-text-to-exp-window :text *response* :x *position* :y 90 :width 25)    
          (proc-display) 
          (log-experiment-event 3)
          (setf *position* (+ *position* 8)))
        (progn
          (delete-event *timeout-event*)
          (schedule-post-answer)))
      ;(schedule-break-relative 0))
      )
    ; handle in n-back here, if anything
    (log-experiment-event 3)
    )
  )

(defun schedule-post-answer ()
  ; T+0: feedback
  ; T+5: first rd letter
  ; T+6: first letter gone
  ; T+6.25: second letter
  ; T+7.25: second letter gone
  ; T+7.5: 3rd letter 
  ; ...
  ; T+16.25: 10th letter appears
  ; T+17: rd over 
  (schedule-event-relative 0 (lambda () ; feedback start
        (log-experiment-event 4)
        (setf *problem-result* (reverse *problem-result*))
        (let ()
          (princ (format nil "Starting feedback~%"))
          (clear-exp-window)
          (present-feedback *problem*))
          (proc-display :clear t)))
  (schedule-event-relative 8 (lambda () ; after feedback
        (princ (format nil "Starting repetition detection~%"))
        (setf *in-nback* t)
        (log-experiment-event 6)
        (clear-exp-window)
        (present-repetition-detection)))
  (loop for i from 1 to 9 do
        (schedule-event-relative (+ 8 (* 1.25 (- i 1))) (lambda ()
                                                        (present-rd-number i)
                                                        (proc-display :clear t)))
        (schedule-event-relative (+ 9 (* 1.25 (- i 1))) (lambda ()
                                                        (clear-rd-number)
                                                        (proc-display :clear t))))
  (schedule-event-relative 19.25 (lambda () 
                                   (present-rd-number 10)
                                   (proc-display :clear t)))
  (schedule-event-relative 20 (lambda () 
                                (setf *in-nback* nil)
                                (push *problem-result* *results*)))
  (schedule-break-relative 20.001)
  nil)


(defun synchronize-fixation ()
  ; scanner waits for two pulses after displaying cross
  (let* ((current-scan (/ (get-time) 2000.0))
        (next-scan-start (ceiling current-scan))
        (diff (- next-scan-start current-scan))
        (wait-time (+ 2 (* 2 diff))))
    (princ (format nil "Waiting ~A seconds to begin on a scan boundary~%" wait-time))
    wait-time))


(defun present-problem (items)
 (cond ((atom (first items))
         (add-text-to-exp-window :text 
                                 (format nil "~A" (first items)) :x 75 :y 60 :width 25))
        (t (add-text-to-exp-window :text 
                                   (format nil "~A" (first (first items))) :x 50 :y 60 :width 25)
           (add-text-to-exp-window :text 
                                   (format nil "~A" (second (first items))) :x 75 :y 60 :width 25)))
  (add-text-to-exp-window :text "$" :x 100 :y 60 :width 25)
  (add-text-to-exp-window :text (format nil "~A" (second items)) :x 125 :y 60 :width 25)
  (add-text-to-exp-window :text "=" :x 150 :y 60 :width 25)
  (cond ((atom (third items)) 
         (add-text-to-exp-window :text (format nil "~A" (third items)) :x 175 :y 60 :width 25))
        (t (add-text-to-exp-window :text (concatenate 'string (format nil "~A" (first (third items))) "_" 
                                                      (format nil "~A"(second (third items)))) :x 175 :y 60 :width 50))))


(defun present-feedback (problem)
  (let ((base (first problem))
        (height (third problem))
        (value (fifth problem))
        (expansion nil))
    (setf expansion (format nil "~A $ ~A = ~A" base height base))
    (loop for i from 2 to height do
          (setf expansion (concatenate 'string 
                                       expansion (format nil " + ~A"
                                                         (- base (- i 1))))))
    (setf expansion (concatenate 'string expansion (format nil " = ~A" value)))
    (print expansion)
    (add-text-to-exp-window :text expansion :x 10 :y 60 :color 'green)
    ))

(defun present-repetition-detection ()
  (add-button-to-exp-window :x 80 :y 155 :height 60 :width 120 :text "Match!")) 


(defun present-rd-number (num)
  (log-experiment-event 5)
  (if (and (> num 1)
           (< (act-r-random 1.0) .25))
    (setf *nback-match* 1)
    (setf *nback-match* 0))
  (setf *rd-letter* (add-text-to-exp-window :text (format nil "~A" num) :x 75 :y 100 :width 25)))

(defun clear-rd-number ()
  (remove-items-from-exp-window *rd-letter*))

(defun log-experiment-event (val)
  (princ (format nil "logging at time ~A~%" (/ (get-time) 1000.0)))
  (push (list (/ (get-time) 1000.0) val) *experiment-events*))




(defmethod rpm-window-key-event-handler ((win rpm-window) key)
  (if (not *in-nback*)
    (if (not (string-equal key "r"))
      (if (not (string-equal key "c"))
        (progn 
        (add-text-to-exp-window :text (string key) :x *position* :y 90 :width 25)    
        (proc-display) 
        (push (list (- (/ (get-time) 1000.0) *problem-start-time*) key) *problem-result*)
        (setf *position* (+ *position* 8)))
    ; handle in n-back here, if anything
        (log-experiment-event 3))
      (progn
        (delete-event *timeout-event*)
        (schedule-post-answer))))
      ;(schedule-break-relative 0)))
  )


(clear-all)


; need to support up to max answer
(defparameter *min-dm-number* -20)
(defparameter *max-dm-number* 40)
(defun add-numbers-to-dm ()
  (loop for i from *min-dm-number* to *max-dm-number* do
        (let ((string-val (format nil "~A" i)))
          (add-dm-fct (list (list (intern (concatenate 'string "N" string-val)) 'isa 'equation-token 'string string-val))))))

(defun set-number-similarities ()
  ; assuming similarities range -1-0
  ; Based on Lebiere (p.48), the similarities are equal to the ratio of the
  ; smaller number to the larger (minus 1 to make them -1 to 0)
  ;
  ; negative numbers don't currently have similarities (all max-difference)
  ; zero is also maximally different from everything else
  (loop for i from 1 to *max-dm-number* do
        (loop for j from (+ i 1) to *max-dm-number* do
              (set-similarities-fct (list (list (number-to-chunk i) (number-to-chunk j)
                                                (- (/ i j) 1)))))))
(defun initialize-position-chunks (factor) 
  ; add some chunks to dm that can be used to index the position in a
  ; problem-solving process
  ;
  ; make them similar based the factor:
  ; similarity = factor*(ratio of smaller step # to larger) - 1
  ;
  ; assuming ms/md of 0/-1, factors should range from 0 (everything
  ; maximally different) to 2 (steps 1 and 2 maximally similar, 1 and 3 at
  ; -1/3, ..)
  ;
  ; going above 2 will cause steps 1 and 2 to have a similarity over 0, so its
  ; probably best to avoid that
  (loop for i from 1 to 40 do
        (let ((string-val (format nil "~A" i)))
          (add-dm-fct (list (list (intern (concatenate 'string "P" string-val)) 'isa 'problem-position)))))
  (loop for i from 1 to 40 do
        (loop for j from (+ i 1) to 40 do
              (set-similarities-fct (list (list (intern (concatenate 'string "P" (format nil "~A" i)))
                                                (intern (concatenate 'string "P" (format nil "~A" j)))
                                                (- (* factor (/ i j)) 1)))))))

(defun next-problem-position (current)
  (intern (concatenate 'string "P" (format nil "~A" (+ 1 (chunk-to-number current))))))


; works on number chunks and problem-positions
(defun chunk-to-number (x)
  (read-from-string (subseq (symbol-name x) 1)))
  
; only works on number chunks
(defun number-to-chunk (x)
  (intern (concatenate 'string "N" (format nil "~A" x))))
  
    
(defun generate-addition-fact (addend1 addend2) 
  (list 'isa 'addition-fact 'addend1 (number-to-chunk addend1) 'addend2 (number-to-chunk addend2) 'sum (number-to-chunk (+ addend1 addend2)))
)

(defun generate-addition-facts (digits1 digits2) 
  (mapcan #'(lambda (x) (mapcar #'(lambda (y) (generate-addition-fact x y)) digits2)) digits1)
)

; adds a set of addition facts to DM, where addend1 takes on each of the values in
; addend1s and addend2 takes on those in addend2s, and assign the new facts the
; given base-level and creation-time
(defun add-addition-facts (addend1s addend2s base-level creation-time)
   (set-base-levels-fct 
     (mapcar #'(lambda(x) (list x base-level creation-time)) 
             (add-dm-fct (generate-addition-facts addend1s addend2s)))))

; this adds addition-facts to dm with their base activation set based on the
; analysis in Lebiere, 1998 (p. 19-20)
; - relative exposure frequency for i+j is rf(i,j)=(2-i/9)(2-j/9)
; - over all problems using 0-9, the sum of all rf's is 225, so the percentage of problems 
; that are i,j is pct=(1/225)rf(i,j)
; - it is assumed that 500,000 problems are presented in these proportions at
; a rate of 100 per day
; - this puts the creation time at (500,000/100)*24*60*60 = 432,000,000s in the
; past
; - since there are 500,000 problems presented, the number of presentations for a
; particular problem is 500,000*(1/225)*rf(i,j), or approx. 2222*rf
;
; the theory does not cover negative numbers, and the model doesn't use them
; much, so they will be given the same activation as their positive counterparts
; 
; this results in activations ranging from -1.55 (for 9+9) to -.163 (for 0+0)
; this may be too low, so a factor can be used to multiply the number of
; references


(defun add-weighted-addition-facts (factor)
  (let ((digits '(-10 -9 -8 -7 -6 -5 -4 -3 -2 -1 0 1 2 3 4 5 6 7 8 9 10 )))
    (mapcan #'(lambda (x) 
                (mapcar #'(lambda (y) (set-base-levels-fct
                                        (mapcar #'(lambda(chunkname) 
                                                    (list chunkname
                                                          (* factor (* 2222 (* (- 2 (/ (abs x) 9.0)) (- 2 (/ (abs y) 9.0)))))
                                                          -432000000))
                                                (add-dm-fct (list (generate-addition-fact x y))))))
                            digits)) digits)))


; these should be replaced with the capability to read digit-by-digit from the
; display


(defun ones (x) (number-to-chunk (mod (chunk-to-number x) 10)))
(defun tens (x) (number-to-chunk (floor (/ (chunk-to-number x) 10))))
(defun combine (x y) (number-to-chunk (+ (* (chunk-to-number x) 10) (chunk-to-number y))))
(defun invert (x) (number-to-chunk (* (chunk-to-number x) -1)))
(defun increment (x) (number-to-chunk (+ (chunk-to-number x) 1)))
(defun decrement (x) (number-to-chunk (+ (chunk-to-number x) -1)))
(defun add (x y) (number-to-chunk (+ (chunk-to-number x) (chunk-to-number y))))
(defun tostring (x) (number-to-chunk x))
(defun issimpleaddition (x y) (let ((num1 (chunk-to-number x))
                             (num2 (chunk-to-number y)))
                      (or (= num1 1)
                          (= num1 0)
                          (= num1 -1)
                          (= num2 1)
                          (= num2 0)
                          (= num2 -1))))

(defun chunk-to-string (x)
  (format nil "~A" (chunk-to-number x)))

(defun string-to-chunk (x)
  (number-to-chunk (read-from-string x)))



(defun parse-and-augment-trace-lists-for-bold (bm)
   (let* ((trace (get-current-buffer-trace))
          (b (bold-module-buffers bm))
          (buffers (if (listp b) b (buffers)))
          (all-data nil))

     (dolist (x buffers) ; for each traced buffer
       (let ((rects nil)
             (current-rect nil)
             (current-request nil))
           (dolist (z trace) ; for each time in the trace
             (let* ((record (find x (buffer-record-buffers z) :key 'buffer-summary-name))
                   (request (buffer-summary-request record))) ; record is the record structure for the buffer at the time
               (if current-rect ; if there is already an event going on
                  (when (or (null (buffer-summary-busy record)) ; and it should end
                             (buffer-summary-busy->free record)
                             (buffer-summary-request record))

                    ;(push (cons current-rect (list (buffer-record-time-stamp z))) rects)  ; add the start->end event
                    (push (cons current-rect (list (buffer-record-time-stamp z) current-request)) rects)  ; add the start->end event
                    (if (buffer-summary-request record) ; and start a new event if necessary
                         ;(setf current-rect (buffer-record-time-stamp z))
                         (progn (setf current-rect (buffer-record-time-stamp z)) (setf current-request request))
                       ;(setf current-rect nil)))
                       (progn (setf current-rect nil) (setf current-request nil))))
                ; current-rect false
                 (if (buffer-summary-busy record) ; if busy,
                   (if (and (buffer-summary-request record) ; if there is a request and a chunk name, or error is set and error->clear is not, or busy->free is set
                            (or (buffer-summary-chunk-name record) 
                                (and (buffer-summary-error record)
                                     (not (buffer-summary-error->clear record)))
                                (buffer-summary-busy->free record)))
                       (push (cons (buffer-record-time-stamp z) (list (buffer-record-time-stamp z) request)) rects) ; add a record starting and ending at this time
                     (progn (setf current-rect (buffer-record-time-stamp z)) (setf current-request request))) ; otherwise save the time stamp in current-rect
                   (if (buffer-summary-request record) ; if not busy, but there is a request,
                       (push (cons (buffer-record-time-stamp z) (list (buffer-record-time-stamp z) request)) rects) ; add a record starting and ending at this time
                     (when (buffer-summary-chunk-name record) ; if not busy and not request, but chunk-name is set, do the same
                       (push (cons (buffer-record-time-stamp z) (list (buffer-record-time-stamp z) request)) rects)))))))

         (push (cons x (reverse rects)) all-data)))

      all-data))


(defun predict-augmented-bold-response (&optional (start 0) (end (mp-time)))  (print 'hi)
  (let* ((bm (get-module bold))
         (data  (parse-and-augment-trace-lists-for-bold bm))
         (bold nil)
         (point (bold-module-point bm))
         (inc (bold-module-inc bm)))
    
    (if (< (- end start) inc)
        (print-warning "Sample time too short for BOLD predictions - must be at least :bold-inc seconds (currently ~s)" inc)
      (progn 
        (unless (zerop (mod start inc))
          (setf start (* inc (floor start inc)))
          (model-warning "Start time should be a multiple of :bold-inc (~S).  Using start time of ~S."
                         inc start))
        
        (dolist (x data)
          (if (find (car x) point)
              (push (cons (car x) (bold_point-predict (mapcar #'car (cdr x)) bm start end)) bold)
            (push (cons (car x) (bold_interval-predict (cdr x) bm start end)) bold)))
        
        ;; Cache the max values for each buffer so that they can
        ;; be used in normalizing things later if desired
        
        ; This can cause errors with long runs, too many args to max
      ; (dolist (x bold)
      ;   (let* ((buffer (car x))
      ;          (data (cdr x))
      ;          (max (when data (apply #'max data))))
      ;     (when (or (null (gethash buffer (bold-module-max-table bm)))
      ;               (> max (gethash buffer (bold-module-max-table bm))))
      ;       (setf (gethash buffer (bold-module-max-table bm)) max))))
        (print 'hi)
        (output-bold-response-data bold bm start end)
        bold))))
